@@ -479,11 +479,16 @@ class XbeImageHeader(ctypes.LittleEndianStructure, StructurePrintMixin):
 		('logo_size',                   ctypes.c_uint32),
 		]
 
-class XbeImageHeaderExtended(XbeImageHeader):
+class XbeImageHeaderExtendedType1(XbeImageHeader):
 	_pack_ = 1
 	_fields_ = [
 		('lib_features_addr',           ctypes.c_uint32),
 		('lib_features_count',          ctypes.c_uint32),
+		]
+
+class XbeImageHeaderExtendedType2(XbeImageHeaderExtendedType1):
+	_pack_ = 1
+	_fields_ = [
 		('debug_info',                  ctypes.c_uint32),
 		]
 
@@ -520,12 +525,16 @@ class XbeImageCertificate(ctypes.LittleEndianStructure, StructurePrintMixin):
 		('version',            ctypes.c_uint32),
 		('lan_key',            ctypes.c_uint8 * 16),
 		('signature_key',      ctypes.c_uint8 * 16),
-		('alt_signature_keys', (ctypes.c_uint8 * (16*16))),
+		('alt_signature_keys', (ctypes.c_uint8 * 16)*16),
 		]
 
 class XbeImageCertificateExtended(XbeImageCertificate):
+	_pack_ = 1
 	_fields_ = [
-		('unknown',            ctypes.c_uint8 * 28), # FIXME
+		('original_certificate_size', ctypes.c_uint32),
+		('online_service_id', ctypes.c_uint32),
+		('security_flags', ctypes.c_uint32),
+		('code_enc_key', ctypes.c_uint8 * 16),
 		]
 
 
@@ -587,6 +596,13 @@ class XbeTlsHeader(ctypes.LittleEndianStructure, StructurePrintMixin):
 		('characteristics',   ctypes.c_uint32),
 		]
 
+class XbeImportDescriptor(ctypes.LittleEndianStructure, StructurePrintMixin):
+	_pack_ = 1
+	_fields_ = [
+		('thunk_array_addr',  ctypes.c_uint32),
+		('image_name_addr',   ctypes.c_uint32),
+		]
+
 class Xbe:
 	ENTRY_DEBUG   = 0x94859D4B
 	ENTRY_RETAIL  = 0xA8FC57AB
@@ -619,8 +635,10 @@ class Xbe:
 		self.header = XbeImageHeader.from_buffer_copy(data, 0)
 		if self.header.image_header_size == ctypes.sizeof(XbeImageHeader):
 			pass
-		elif self.header.image_header_size == ctypes.sizeof(XbeImageHeaderExtended):
-			self.header = XbeImageHeaderExtended.from_buffer_copy(data, 0)
+		elif self.header.image_header_size == ctypes.sizeof(XbeImageHeaderExtendedType1):
+			self.header = XbeImageHeaderExtendedType1.from_buffer_copy(data, 0)
+		elif self.header.image_header_size == ctypes.sizeof(XbeImageHeaderExtendedType2):
+			self.header = XbeImageHeaderExtendedType2.from_buffer_copy(data, 0)
 		else:
 			log.warning("Unexpected XBE image header size!")
 			assert(self.header.image_header_size > ctypes.sizeof(XbeImageHeader))
@@ -686,7 +704,7 @@ class Xbe:
 			sec_data_start = sec_hdr.raw_addr
 			sec_data_end = sec_data_start + sec_hdr.raw_size
 			sec_data = data[sec_data_start:sec_data_end]
-			self.sections[sec_name] = XbeSection(sec_name, sec_hdr, sec_data)
+			self.sections[sec_name] = XbeSection(sec_name_base, sec_hdr, sec_data)
 
 			log.debug(('Section %d: %s\n' % (i, sec_name)) + sec_hdr.dumps(indent=2))
 			sec_hdr_offset += ctypes.sizeof(XbeSectionHeader)
@@ -695,7 +713,7 @@ class Xbe:
 		cert_offset = self.vaddr_to_file_offset(self.header.certificate_addr)
 		log.debug('Parsing image certificate at offset 0x%x' % cert_offset)
 		# FIXME: Validate address
-		self.cert = XbeImageCertificate.from_buffer_copy(data, cert_offset)		
+		self.cert = XbeImageCertificate.from_buffer_copy(data, cert_offset)
 		if self.cert.size == ctypes.sizeof(XbeImageCertificate):
 			pass
 		elif self.cert.size == ctypes.sizeof(XbeImageCertificateExtended):
@@ -723,7 +741,7 @@ class Xbe:
 				lib_ver_offset += ctypes.sizeof(XbeLibraryVersion)
 
 		# Parse library features
-		if isinstance(self.header, XbeImageHeaderExtended) and self.header.lib_features_addr != 0:
+		if isinstance(self.header, XbeImageHeaderExtendedType1) and self.header.lib_features_addr != 0:
 			lib_feat_offset = self.vaddr_to_file_offset(self.header.lib_features_addr)
 			log.debug('Parsing library features at offset 0x%x' % lib_feat_offset)
 			for i in range(self.header.lib_features_count):
@@ -742,20 +760,48 @@ class Xbe:
 
 		# Parse kernel imports
 		# FIXME: Validate address
-		kern_thunk_table_offset = self.header.kern_thunk_addr
-		kern_thunk_table_offset ^= Xbe.KTHUNK_DEBUG if self.is_debug else Xbe.KTHUNK_RETAIL
-		kern_thunk_table_offset = self.vaddr_to_file_offset(kern_thunk_table_offset)
-		log.debug('Parsing kernel thunk table at offset 0x%x' % kern_thunk_table_offset)
+		kern_thunk_table_addr = self.header.kern_thunk_addr
+		kern_thunk_table_addr ^= Xbe.KTHUNK_DEBUG if self.is_debug else Xbe.KTHUNK_RETAIL
+		kern_thunk_table_offset = self.vaddr_to_file_offset(kern_thunk_table_addr)
+		log.debug('Parsing kernel thunk table at address %x (offset %#x)', kern_thunk_table_addr, kern_thunk_table_offset)
 		i = 0
 		while True:
 			x = ctypes.c_uint32.from_buffer_copy(data, kern_thunk_table_offset)
-			if x.value == 0: break
+			if x.value == 0:
+				break
 			import_name = XbeKernelImage.exports[x.value-0x80000000]
 			self.kern_imports.append(import_name)
-			log.debug('Import %d: 0x%x - %s' % (i, x.value, import_name))
+			thunk_addr = kern_thunk_table_addr + i * 4
+			log.debug('Import %d: 0x%x - %s @ %x' % (i, x.value, import_name, thunk_addr))
 			i += 1
 			kern_thunk_table_offset += 4
 
+		# Parse imports
+		self.imports = []
+		if self.header.import_dir_addr != 0:
+			import_dir_offset = self.vaddr_to_file_offset(self.header.import_dir_addr)
+			log.debug('Parsing kernel imports at address %x (offset %#x)', self.header.import_dir_addr, import_dir_offset)
+			while True:
+				entry = XbeImportDescriptor.from_buffer_copy(data, import_dir_offset)
+				if entry.thunk_array_addr == 0 or entry.image_name_addr == 0:
+					break
+				image_name = self.get_wcstring_from_offset(data, self.vaddr_to_file_offset(entry.image_name_addr))
+				self.imports.append((image_name, entry.thunk_array_addr))
+				log.debug('XBE imports image %s (name at %x)', image_name, entry.image_name_addr)
+				thunk_array_offset = self.vaddr_to_file_offset(entry.thunk_array_addr)
+				thunk_addr = entry.thunk_array_addr
+				while True:
+					x = ctypes.c_uint32.from_buffer_copy(data, thunk_array_offset)
+					if x.value == 0:
+						break
+
+					ordinal = x.value & 0x7fffffff
+					log.debug('[%s] Import %d @ %x', image_name, ordinal, thunk_addr)
+					thunk_array_offset += 4
+					thunk_addr += 4
+
+				import_dir_offset += ctypes.sizeof(XbeImportDescriptor)
+
 	def get_cstring_from_offset(self, data, offset, encoding=None):
 		"""Read null-terminated string from `offset` in `data`"""
 		name = bytearray()
@@ -803,14 +849,15 @@ class Xbe:
 		def off_to_addr(off):
 			return self.header.base_addr + off
 
+		# FIXME: Adjust offset dynamically, don't depend on existing slack
 		raw_off = round_up(self.header.headers_size)
 
 		# Construct section data
 		section_data = bytes()
 		for name in sorted(self.sections, key=lambda x: self.sections[x].header.virtual_addr):
 			s = self.sections[name]
-			print(name)
-			print("EXPECTED %8x GOT %8x" % (s.header.raw_addr, raw_off))
+			if s.header.raw_addr != 0 and s.header.raw_addr != raw_off:
+				log.warning("Expected %8x for %s, got %8x" % (s.header.raw_addr, name, raw_off))
 			# assert(s.header.raw_addr == raw_off)
 			s.header.raw_addr = raw_off
 			section_data += s.data
@@ -872,6 +919,28 @@ class Xbe:
 			self.sections[name].header.section_name_addr = addr
 		do_append(bytes(round_up(raw_off, 4)-raw_off)) # Align to 4 bytes
 
+		# Import directory
+		if self.imports:
+			import_table_offset = raw_off
+			import_table_size = ctypes.sizeof(XbeImportDescriptor)*(len(self.imports) + 1)
+			import_strings_offset = import_table_offset + import_table_size
+			for name, addr in self.imports:
+				desc = XbeImportDescriptor()
+				desc.thunk_array_addr = addr
+				desc.image_name_addr = off_to_addr(import_strings_offset)
+				import_strings_offset += 2*(len(name)+1)
+				do_append(bytes(desc))
+
+			desc = XbeImportDescriptor()
+			do_append(bytes(desc))
+			self.header.import_dir_addr = off_to_addr(import_table_offset)
+
+		# Import names
+		for name, addr in self.imports:
+			do_append(name.encode('utf_16_le') + b'\x00\x00')
+		if self.imports:
+			do_append(b'\x00\x00')  # ???
+
 		# Library versions
 		self.header.lib_versions_addr = off_to_addr(raw_off)
 		self.header.lib_versions_count = len(self.libraries)
@@ -885,7 +954,7 @@ class Xbe:
 				self.header.xapi_lib_version_addr = addr
 
 		# Library features
-		if isinstance(self.header, XbeImageHeaderExtended) and len(self.library_features) > 0:
+		if isinstance(self.header, XbeImageHeaderExtendedType1) and len(self.library_features) > 0:
 			self.header.lib_features_count = len(self.library_features)
 			self.header.lib_features_addr = off_to_addr(raw_off)
 			for l in self.library_features:
@@ -906,20 +975,18 @@ class Xbe:
 		# Additional fixups
 
 		# Sometimes this includes padding, other times it does not, what gives?
-		# self.header.headers_size = raw_off
-		do_append(bytes(round_up(raw_off)-raw_off)) # Align to 4K
+		do_append(bytes(round_up(raw_off, 0x2000)-raw_off)) # Align to 4K
+		self.header.headers_size = raw_off
 
-		print('HEADERS SIZE = %x\n' % self.header.headers_size)
-		print('RAW_OFF = %x\n' % raw_off)
+		self.header.image_size = max(s.header.virtual_size+s.header.virtual_addr for s in self.sections.values()) - self.header.base_addr
 
 		# Construct final image
 		output = bytes()
 		output += bytes(self.header)
 		output += bytes(self.cert)
-		for name in self.sections:
-			s = self.sections[name].header
-			print("writing section at %x" % len(output))
-			output += bytes(s)
+		for name, s in self.sections.items():
+			log.info("Writing section %s header at %x" % (name, len(output)))
+			output += bytes(s.header)
 		output += headers_data
 		output += section_data
 
@@ -927,9 +994,7 @@ class Xbe:
 		# FIXME: Why?
 		if output[-1] != 0:
 			output += bytes(0x1000)
-
-		with open('out.xbe', 'wb') as f:
-			f.write(output)
+		return output
 
 	@classmethod
 	def from_file(cls, path):
@@ -1077,6 +1142,75 @@ def decode_xpr_image(data):
 
 	return (w, h, decode_bc1(w, h, data[hdr.header_size:]))
 
+def encode_logo(pixels):
+	"""
+	Encode pixel data into the RLE-compressed logo format
+	"""
+	assert(len(pixels) == 1700, "Invalid logo size")
+
+	raw = [round((pixel[0] + pixel[1] + pixel[2]) * 255 / 3) for pixel in pixels]
+
+	encoded = []
+
+	i = 0
+	while i < len(raw):
+		color = raw[i] >> 4
+		length = 1
+
+		while i+1 < len(raw) and length < 1023 and (raw[i+1] >> 4) == color:
+			length += 1
+			i += 1
+
+		if length < 8:
+			# Type 1
+			encoded.append(
+				0x01 |
+				(length << 1) |
+				((color & 0x0F) << 4)
+			)
+		else:
+			# Type 2
+			value = 0x02 | \
+					(length << 2) | \
+					((color & 0x0F) << 12)
+			encoded.extend(struct.pack("<H", value))
+
+		i += 1
+
+	return bytes(encoded)
+
+def decode_logo(data):
+	"""
+	Decode an XBE logo (RLE-compressed)
+	"""
+	w = 100
+	h = 17
+	pixels = [(0,0,0,1) for _ in range(w*h)]
+	i = 0
+	c = 0
+
+	while i < len(data):
+		if data[i] & 0x01:
+			# Type 1
+			length = get_bits(data[i], 4-1, 1)
+			byte = (get_bits(data[i], 8-1, 4) << 4) / 255
+			for j in range(0, length):
+				pixels[c] = (byte, byte, byte, 1)
+				c += 1
+		else:
+			# Type 2
+			d = struct.unpack("<H", data[i:i+2])[0]
+			assert(d & 0x0002 == 0, "Invalid type2 value")
+			length = get_bits(d, 12-1, 2)
+			byte = (get_bits(d, 16-1, 12) << 4) / 255
+			i += 1
+			for j in range(0, length):
+				pixels[c] = (byte, byte, byte, 1)
+				c += 1
+		i += 1
+
+	return (w, h, pixels)
+
 def encode_bmp(w, h, pixels):
 	"""
 	Encode a standard Windows BMP Image File
